# encoding: utf-8

# Lista de menús dietéticos.
module Linkedlist
    
    # Plato de un menú.
    class Plato
        
        # Descripción del plato.
        attr_reader :descr
        # Tamaño de la porción recomendada.
        attr_reader :porcion
        # Ingesta en gramos.
        attr_reader :ingesta
        
        # Inicializa el plato con los datos pasados.
        def initialize(descr, porcion, ingesta)
            raise ArgumentError, "[Plato.initialize] La ingesta en gramos debe ser mayor que 0" unless ingesta > 0
            @descr, @porcion, @ingesta = descr, porcion, ingesta 
        end
        
        # Comprobación de la igualdad entre platos.
        def ==(plato)
            return @descr == plato.descr && @porcion == plato.porcion && @ingesta == plato.ingesta
        end
        
        # Comprobación de la igualdad entre platos.
        def eql?(plato)
            return self == plato
        end
        
        # Devuelve el plato formateado como cadena.
        def to_s
           return "#{@descr}, #{@porcion}, #{@ingesta} g" 
        end
        
    end
    
    # Menú básico.
    class Menu
        
        # Incluye Comparable
		include Comparable
        
        # Título del menú.
        attr_reader :titulo
        # Porcentaje de ingesta diaria.
        attr_reader :pct_ingesta
        # Valor calórico total (VCT) en kcal.
        attr_reader :vct
        # Porcentaje de proteínas de sus platos.
        attr_reader :pct_prot
        # Porcentaje de grasas de sus platos.
        attr_reader :pct_grasas
        # Porcentaje de hidratos de carbono de sus platos.
        attr_reader :pct_hdc
        # Array de platos del menú.
        attr_reader :platos
        
        # Inicializa el menú con los datos pasados (formato DSL).
        def initialize(&block)
            instance_eval(&block)
            raise ArgumentError, Menu.err_pct('ingesta diaria') unless Menu.is_pct?(@pct_ingesta)
            raise ArgumentError, Menu.err_pct('proteínas') unless Menu.is_pct?(@pct_prot)
            raise ArgumentError, Menu.err_pct('grasas') unless Menu.is_pct?(@pct_grasas)
            raise ArgumentError, Menu.err_pct('hidratos') unless Menu.is_pct?(@pct_hdc)
            raise ArgumentError, "[Menu.initialize] Los porcentajes de macronutrientes deben sumar 100%" unless @pct_prot + @pct_grasas + @pct_hdc == 100
            raise ArgumentError, "[Menu.initialize] Debe haber al menos un plato" unless @platos.count > 0
        end
        
        # Establece el título del menú.
        def tit(titulo)
           @titulo = titulo
        end
        
        # Establece el porcentaje de ingesta diaria.
        def ingesta(ingesta)
           @pct_ingesta = ingesta
        end
        
        # Establece el valor calórico total (VCT) y los porcentajes de proteínas, grasas e hidratos de carbono.
        def porcentajes(porcentajes)
           @vct = porcentajes[:vct]
           @pct_prot = porcentajes[:proteinas]
           @pct_grasas = porcentajes[:grasas]
           @pct_hdc = porcentajes[:hidratos]
        end
        
        # Añade un plato al menú.
        def plato(plato)
           @platos = [] if @platos.nil?
           @platos.push(Plato.new(plato[:descripcion], plato[:porcion], plato[:gramos]))
        end
        
        # Obtiene el plato correspondiente al número _num_ o nil si éste no existe.
        def get_plate(num)
           return (num >= 0 && num < platos.count) ? platos[num] : nil
        end
        
        # Devuelve el menú formateado como cadena.
        def to_s
           s = ''
           s << "#{@titulo.upcase} (#{@pct_ingesta}%)\n" # Título y porcentaje de ingesta diaria
           @platos.each {|plato| s << "- #{plato.to_s}\n"} # Platos
           s << "V.C.T. | % #{@vct} kcal | #{@pct_prot}% - #{pct_grasas}% - #{pct_hdc}%\n" # VCT y porcentajes de macronutrientes
           return s
        end
        
        # Comprueba que _num_ está entre 0 y 100 (ambos incluidos).
        def self.is_pct?(num)
           return (0.0..100.0).include?(num)
        end
        
        # Devuelve la cadena de error si el porcentaje no es correcto.
        def self.err_pct(str)
           return "[Menu.initialize] El porcentaje de #{str} debe estar entre 0 y 100" 
        end
        
        # Comparación de menús.
        # La comparación es por orden alfabético de títulos y por cantidad de platos (de más a menos) si hay igualdad en los títulos.
		def <=>(c_menu)
			titulos_ord = @titulo <=> c_menu.titulo
			if titulos_ord == 0
			   return -(@platos.count <=> c_menu.platos.count)
			else
			    return titulos_ord
			end
		end
        
    end
    
    # Menú por grupo de alimentos.
    class MenuGrupos < Menu
    	
    	# Grupo de alimentos.
    	attr_reader :grupo_alimentos
    	
    	# Inicializa el menú con los datos pasados (formato DSL).
    	def initialize(&block)
    		instance_eval(&block)
    		raise ArgumentError, Menu.err_pct('ingesta diaria') unless Menu.is_pct?(@pct_ingesta)
            raise ArgumentError, Menu.err_pct('proteínas') unless Menu.is_pct?(@pct_prot)
            raise ArgumentError, Menu.err_pct('grasas') unless Menu.is_pct?(@pct_grasas)
            raise ArgumentError, Menu.err_pct('hidratos') unless Menu.is_pct?(@pct_hdc)
            raise ArgumentError, "[Menu.initialize] Los porcentajes de macronutrientes deben sumar 100%" unless @pct_prot + @pct_grasas + @pct_hdc == 100
            raise ArgumentError, "[Menu.initialize] Debe haber al menos un plato" unless @platos.count > 0
    	end
    	
    	# Establece el grupo de alimentos.
    	def grupo(grupo)
            @grupo_alimentos = grupo 
    	end
    	
    	# Devuelve el menú por grupo de alimentos formateado como cadena.
    	def to_s
    		return super << "El menú está formado exclusivamente por #{@grupo_alimentos}\n" # Grupo de alimentos
    	end
    	
    end
    
    # Menú por edades.
    class MenuEdades < Menu
    	
    	# Rango de edades recomendado para el menú.
    	attr_reader :rango_edad
    	
    	# Inicializa el menú con los datos pasados (formato DSL).
    	def initialize(&block)
    		instance_eval(&block)
    		raise ArgumentError, Menu.err_pct('ingesta diaria') unless Menu.is_pct?(@pct_ingesta)
            raise ArgumentError, Menu.err_pct('proteínas') unless Menu.is_pct?(@pct_prot)
            raise ArgumentError, Menu.err_pct('grasas') unless Menu.is_pct?(@pct_grasas)
            raise ArgumentError, Menu.err_pct('hidratos') unless Menu.is_pct?(@pct_hdc)
            raise ArgumentError, "[Menu.initialize] Los porcentajes de macronutrientes deben sumar 100%" unless @pct_prot + @pct_grasas + @pct_hdc == 100
            raise ArgumentError, "[Menu.initialize] Debe haber al menos un plato" unless @platos.count > 0
    		raise ArgumentError, "[MenuEdades.initialize] El rango de edades debe ser de tipo Range" unless @rango_edad.kind_of?(Range)
    	end
    	
    	# Establece el rango de edades recomendado para el menú.
    	def rango(rango)
    	    @rango_edad = rango
    	end
    	
    	# Devuelve el menú por edades formateado como cadena.
    	def to_s
    		return super << "Apropiado para edades comprendidas entre los #{@rango_edad.first} y los #{@rango_edad.last} años\n" # Rango de edades
    	end
    	
    end
    
    # Nodo que contiene un menú dietético de cualquier tipo.
    class Node

		# Valor del nodo (menú dietético).
		attr_accessor :value
		# Enlace al siguiente nodo de la lista de menús dietéticos.
		attr_accessor :next_node
		# Enlace al anterior nodo de la lista de menús dietéticos.
		attr_accessor :prev_node
  
    	# Inicializa el nodo con el valor _value_. Si se especifican también inicializa los enlaces al siguiente y anterior nodo de la lista.
    	def initialize(value, next_node = nil, prev_node = nil)
	        @value = value
	        @next_node = next_node
	        @prev_node = prev_node
    	end
    
    end
    
    # Lista de menús dietéticos.
    class List

        # Incluye Enumerable
		include Enumerable
        
        # Nodo cabeza de la lista.
		attr_reader :head
		# Nodo cola de la lista.
		attr_reader :tail
		# Tamaño de la lista.
		attr_reader :size

        # Inicializa la lista vacía.
        def initialize()
			@head = nil
			@tail = nil
			@size = 0
        end
		
		# Devuelve si la lista está vacía.
		def empty?()
			return (@head == nil)
		end
		
		# Inserta un menú en la lista vacía.
		def push_empty(ref)
			
			raise RuntimeError, "[List.push_empty]: Lista no vacía" unless empty?()
			
			nodo = Node.new(ref)
			@head = nodo
			@tail = nodo
			@size += 1
			
			return self
			
		end
		
		# Inserta el menú _ref_ al final de la lista.
		def push(ref)
			if empty?()
				return push_empty(ref)
			else
				nodo = Node.new(ref, nil, @tail)
				@tail.next_node = nodo
				@tail = nodo
				@size += 1
				return self
			end
		end
		
		# Inserta múltiples elementos al final de la lista.
		def push_multi(*refs)
			refs.each { |ref| push(ref)}
		end
		
		# Extrae y devuelve el menú del principio de la lista.
		def pop()
			
			raise RuntimeError, "[List.pop]: No se puede extraer elementos de una lista vacía" if empty?()
			
			nodo = @head
			if @head.equal?(@tail)
				@head = nil
				@tail = nil
			else
				@head = @head.next_node
				@head.prev_node = nil
			end
			
			@size -= 1
			
			return nodo.value
			
		end
		
		# Itera la lista, ejecutando el bloque pasado pasándole como parámetro el valor de cada nodo.
		def each
			
			nodo = @head
			until nodo.nil?
				yield nodo.value
				nodo = nodo.next_node
			end
			
		end
		
		private :push_empty

    end

end
